import math
import os
import pathlib
import sys
import tarfile
import tempfile

from appdirs import user_data_dir
import numpy as np
from scipy.interpolate import griddata

from pyschism.forcing.tides.base import TidalDataProvider

DATADIR = pathlib.Path(user_data_dir()) / 'tpxo'
DATADIR.mkdir(exist_ok=True, parents=True)


def polar(z):
    a = z.real
    b = z.imag
    r = math.hypot(a, b)
    theta = math.atan2(b, a)
    return r, np.rad2deg(theta)


class TPXO(TidalDataProvider):

    def __init__(self):

        _tarfile = os.getenv('TPXO_NETCDF_TARFILE')
        if _tarfile is not None:
            _tarfile = pathlib.Path(_tarfile)
        else:
            _tarfile = DATADIR / 'tpxo9.tar.gz'

        if not _tarfile.is_file():
            raise FileNotFoundError(
                'No TPXO file found.\nUsers interested in using the TPXO data '
                'as forcing source for the model will need to register and '
                'request a copy of the TPXO9 binary files (specifically '
                'the binary version tpxo9.tar.gz) from the authors at '
                'https://www.tpxo.net. Once you obtain this copy, you can set '
                'you can set the environment variable TPXO_NETCDF_TARFILE '
                'to point to the path of the tpxo9.tar.gz file or you may '
                f'symlink this file manually to {_tarfile.resolve()} path.')

        self._tarfile = _tarfile

        self._gfile = pathlib.Path(self._tmpdir.name) / 'DATA/grid_tpxo9'

        self._hfile = pathlib.Path(self._tmpdir.name) / 'DATA/h_tpxo9.v1'

        self._ufile = pathlib.Path(self._tmpdir.name) / 'DATA/u_tpxo9.v1'

        self.x, self.y, _, mask = read_tide_grid(self._gfile)[:4]
        self.mask = ~mask.astype(bool)

    def get_elevation(self, constituent, vertices):
        index = self.constituents.index(constituent.lower().capitalize())
        array = read_elevation_file(self._hfile, index)
        xq = np.asarray(
            [x + 360. if x < 0. else x for x in vertices[:, 0]]).flatten()
        yq = vertices[:, 1].flatten()
        xi, yi = np.meshgrid(self.x, self.y)
        dx = np.mean(np.diff(self.x))
        dy = np.mean(np.diff(self.y))
        xi = xi.flatten()
        yi = yi.flatten()
        idxq = np.where(
                np.logical_and(  # buffer the bbox by 2 difference units
                    np.logical_and(xi >= np.min(xq) - 2 * dx,
                                   xi <= np.max(xq) + 2 * dx),
                    np.logical_and(yi >= np.min(yq) - 2 * dy,
                                   yi <= np.max(yq) + 2 * dy)
                    )
                )
        values = np.ma.zeros((len(vertices),), dtype=array.dtype)
        values.data.real = griddata((xi[idxq], yi[idxq]),
                                    array.real.flatten()[idxq],
                                    (xq, yq), method='nearest')
        values.data.imag = griddata((xi[idxq], yi[idxq]),
                                    array.imag.flatten()[idxq],
                                    (xq, yq), method='nearest')
        amp, phase = [list(data) for data in zip(*map(polar, values))]
        phase = np.abs(np.array(phase)-360)
        idx = np.where(phase >= 360)
        phase[idx] = phase[idx] % 360
        return amp, phase

    def get_velocity(self, constituent, vertices):
        index = self.constituents.index(constituent.lower().capitalize())
        u, v = read_transport_file(self._ufile, index)
        xq = np.asarray([x + 360. for x in vertices[:, 0] if x < 0]).flatten()
        yq = vertices[:, 1].flatten()
        xi, yi = np.meshgrid(self.x, self.y)
        dx = np.mean(np.diff(self.x))
        dy = np.mean(np.diff(self.y))
        xi = xi.flatten()
        yi = yi.flatten()
        idxq = np.where(
                np.logical_and(  # buffer the bbox by 2 difference units
                    np.logical_and(xi >= np.min(xq) - 2 * dx,
                                   xi <= np.max(xq) + 2 * dx),
                    np.logical_and(yi >= np.min(yq) - 2 * dy,
                                   yi <= np.max(yq) + 2 * dy)
                    )
                )
        uval = np.ma.zeros((len(vertices),), dtype=u.dtype)
        vval = np.ma.zeros((len(vertices),), dtype=v.dtype)

        uval.data.real = griddata((xi[idxq], yi[idxq]),
                                  u.real.flatten()[idxq],
                                  (xq, yq), method='nearest')
        uval.data.imag = griddata((xi[idxq], yi[idxq]),
                                  u.imag.flatten()[idxq],
                                  (xq, yq), method='nearest')
        vval.data.real = griddata((xi[idxq], yi[idxq]),
                                  v.real.flatten()[idxq],
                                  (xq, yq), method='nearest')
        vval.data.imag = griddata((xi[idxq], yi[idxq]),
                                  v.imag.flatten()[idxq],
                                  (xq, yq), method='nearest')

        uamp, uphase = [list(data) for data in zip(*map(polar, uval))]
        uphase = np.abs(np.array(uphase)-360)
        uidx = np.where(uphase >= 360)
        uphase[uidx] = uphase[uidx] % 360

        vamp, vphase = [list(data) for data in zip(*map(polar, vval))]
        vphase = np.abs(np.array(vphase)-360)
        vidx = np.where(vphase >= 360)
        vphase[vidx] = vphase[vidx] % 360
        return uamp, uphase, vamp, vphase

    @property
    def constituents(self):
        return [c.capitalize() for c in read_constituents(self._hfile)[0]]

    @property
    def _tarfile(self):
        return self.__tarfile

    @_tarfile.setter
    def _tarfile(self, file: pathlib.Path):
        self._tmpdir = tempfile.TemporaryDirectory()
        tmpdir = pathlib.Path(self._tmpdir.name)
        with tarfile.open(file) as f:
            f.extractall(path=tmpdir)
        self.__tarfile = file


def install():
    prefix = "/".join(sys.executable.split('/')[:-2])
    file = pathlib.Path(prefix) / 'lib/h_tpxo9.v1.nc'
    TPXO._fetch_tpxo_file(prefix, file)


# https://github.com/tsutterley/pyTMD/blob/47ff266e83b92f0ed75f2710fdedde3f9510bb84/pyTMD/read_tide_model.py#L561-L582
def read_constituents(input_file):
    """
    Read the list of constituents from an elevation or transport file
    Arguments
    ---------
    input_file: input tidal file
    Returns
    -------
    constituents: list of tidal constituent IDs
    nc: number of constituents
    """
    #-- open the file
    fid = open(os.path.expanduser(input_file),'rb')
    ll, = np.fromfile(fid, dtype=np.dtype('>i4'), count=1)
    nx,ny,nc = np.fromfile(fid, dtype=np.dtype('>i4'), count=3)
    fid.seek(16,1)
    constituents = [c.decode("utf-8").rstrip() for c in fid.read(nc*4).split()]
    fid.close()
    return (constituents,nc)


# https://github.com/tsutterley/pyTMD/blob/47ff266e83b92f0ed75f2710fdedde3f9510bb84/pyTMD/read_tide_model.py#L586-L619
def read_elevation_file(input_file,ic):

    """
    Read elevation file to extract real and imaginary components for constituent
    Arguments
    ---------
    input_file: input elevation file
    ic: index of consituent
    Returns
    -------
    h: tidal elevation
    """
    #-- open the file
    fid = open(os.path.expanduser(input_file),'rb')
    ll, = np.fromfile(fid, dtype=np.dtype('>i4'), count=1)
    nx,ny,nc = np.fromfile(fid, dtype=np.dtype('>i4'), count=3)
    #-- extract x and y limits
    ylim = np.fromfile(fid, dtype=np.dtype('>f4'), count=2)
    xlim = np.fromfile(fid, dtype=np.dtype('>f4'), count=2)
    #-- skip records to constituent
    nskip = ic*(nx*ny*8+8) + 8 + ll - 28
    fid.seek(nskip,1)
    #-- real and imaginary components of elevation
    h = np.ma.zeros((ny,nx),dtype=np.complex64)
    h.mask = np.zeros((ny,nx),dtype=np.bool)
    for i in range(ny):
        temp = np.fromfile(fid, dtype=np.dtype('>f4'), count=2*nx)
        h.data.real[i,:] = temp[0:2*nx-1:2]
        h.data.imag[i,:] = temp[1:2*nx:2]
    #-- close the file
    fid.close()
    #-- return the elevation
    return h


# https://github.com/tsutterley/pyTMD/blob/47ff266e83b92f0ed75f2710fdedde3f9510bb84/pyTMD/read_tide_model.py#L712-L752
def read_transport_file(input_file,ic):
    """
    Read transport file to extract real and imaginary components for constituent
    Arguments
    ---------
    input_file: input transport file
    ic: index of consituent
    Returns
    -------
    u: zonal tidal transport
    v: meridional zonal transport
    """
    #-- open the file
    fid = open(os.path.expanduser(input_file),'rb')
    ll, = np.fromfile(fid, dtype=np.dtype('>i4'), count=1)
    nx,ny,nc = np.fromfile(fid, dtype=np.dtype('>i4'), count=3)
    #-- extract x and y limits
    ylim = np.fromfile(fid, dtype=np.dtype('>f4'), count=2)
    xlim = np.fromfile(fid, dtype=np.dtype('>f4'), count=2)
    #-- skip records to constituent
    nskip = ic*(nx*ny*16+8) + 8 + ll - 28
    fid.seek(nskip,1)
    #-- real and imaginary components of transport
    u = np.ma.zeros((ny,nx),dtype=np.complex64)
    u.mask = np.zeros((ny,nx),dtype=np.bool)
    v = np.ma.zeros((ny,nx),dtype=np.complex64)
    v.mask = np.zeros((ny,nx),dtype=np.bool)
    for i in range(ny):
        temp = np.fromfile(fid, dtype=np.dtype('>f4'), count=4*nx)
        u.data.real[i,:] = temp[0:4*nx-3:4]
        u.data.imag[i,:] = temp[1:4*nx-2:4]
        v.data.real[i,:] = temp[2:4*nx-1:4]
        v.data.imag[i,:] = temp[3:4*nx:4]
    #-- close the file
    fid.close()
    #-- return the transport components
    return (u,v)


def read_tide_grid(input_file):
    """
    Read grid file to extract model coordinates, bathymety, masks and indices
    Arguments
    ---------
    input_file: input grid file
    Returns
    -------
    x: x-coordinates of input grid
    y: y-coordinates of input grid
    hz: model bathymety
    mz: land/water mask
    iob: open boundary index
    dt: time step
    """
    #-- open the file
    fid = open(os.path.expanduser(input_file),'rb')
    fid.seek(4,0)
    #-- read data as big endian
    #-- get model dimensions and limits
    nx, = np.fromfile(fid, dtype=np.dtype('>i4'), count=1)
    ny, = np.fromfile(fid, dtype=np.dtype('>i4'), count=1)
    #-- extract x and y limits (these could be latitude and longitude)
    ylim = np.fromfile(fid, dtype=np.dtype('>f4'), count=2)
    xlim = np.fromfile(fid, dtype=np.dtype('>f4'), count=2)
    dt, = np.fromfile(fid, dtype=np.dtype('>f4'), count=1)
    #-- convert longitudinal limits (if x == longitude)
    if (xlim[0] < 0) & (xlim[1] < 0) & (dt > 0):
        xlim += 360.0
    #-- create x and y arrays arrays (these could be lon and lat values)
    dx = (xlim[1] - xlim[0])/nx
    dy = (ylim[1] - ylim[0])/ny
    x = np.linspace(xlim[0]+dx/2.0,xlim[1]-dx/2.0,nx)
    y = np.linspace(ylim[0]+dy/2.0,ylim[1]-dy/2.0,ny)
    #-- read nob and iob from file
    nob, = np.fromfile(fid, dtype=np.dtype('>i4'), count=1)
    if (nob == 0):
        fid.seek(20,1)
        iob = []
    else:
        fid.seek(8,1)
        iob=np.fromfile(fid, dtype=np.dtype('>i4'), count=2*nob).reshape(nob,2)
        fid.seek(8,1)
    #-- read hz matrix
    hz = np.fromfile(fid, dtype=np.dtype('>f4'), count=nx*ny).reshape(ny,nx)
    fid.seek(8,1)
    #-- read mz matrix
    mz = np.fromfile(fid, dtype=np.dtype('>i4'), count=nx*ny).reshape(ny,nx)
    #-- close the file
    fid.close()
    #-- return values
    return (x,y,hz,mz,iob,dt)



#! /usr/bin/env python
import argparse

from pyschism.cmd.forecast.forecast import ForecastCli, add_forecast
from pyschism.cmd.bctides import BctidesCli, add_bctides


def parse_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='mode')
    # add_forecastd(subparsers)
    add_forecast(subparsers)
    # add_plot(subparsers)
    add_bctides(subparsers)
    # add_viewerd(subparsers)
    # add_autodocd(subparsers)
    return parser.parse_args()


def main():
    args = parse_args()

    if args.mode == 'forecast':
        ForecastCli(args)

    elif args.mode == 'bctides':
        BctidesCli(args)

    else:
        raise NotImplementedError(f'Unhandled CLI mode: {args.mode}')


if __name__ == '__main__':
    main()



from pyschism.outputs.plot import PlotOutputCombined
from pyschism.outputs.stations import StationsOutput
from pyschism.outputs.stations.plot import PlotOutputStations

__all__ = [
    'PlotOutputCombined',
    'PlotOutputStations',
    "StationsOutput"
]



from argparse import Namespace
from collections import namedtuple
from datetime import timedelta
import json
import os
# import logging
import subprocess

from pyschism.cmd.forecast.init import ForecastInit
from pyschism.driver import ModelDriver
from pyschism.domain import ModelDomain
from pyschism.forcing.atmosphere.nws.nws2 import NWS2
from pyschism.forcing.atmosphere import GlobalForecastSystem as GFS
from pyschism.forcing.hydrology import NationalWaterModel as NWM
from pyschism.param.schout import SurfaceOutputVars

# logger = logging.getLogger()


class HotstartDirectory:

    def __get__(self, obj, val):
        hotstart_directory = obj.__dict__.get('hotstart_directory')
        if hotstart_directory is None:
            timestamp = str(obj.target_datetime).replace(' ', 'T')
            hotstart_directory = obj.forecasts_directory / f'{timestamp}'
            hotstart_directory.parent.mkdir(exist_ok=True)
            hotstart_directory.mkdir(exist_ok=True)
            obj.__dict__['hotstart_directory'] = hotstart_directory
        return hotstart_directory


class HotstartDomain:

    def __get__(self, obj, val):
        hotstart_domain = obj.__dict__.get('hotstart_domain')
        if hotstart_domain is None:
            # obj.logger.debug('Generating hotstart ModelDomain.')
            hotstart_domain = ModelDomain.open(
                obj.hgrid_path,
                obj.fgrid_path,
                # TODO: Vgrid.open() raises NotImplementedError
                # obj.vgrid_path,
                hgrid_crs=obj.args.hgrid_crs,
                fgrid_crs=obj.args.fgrid_crs
            )
            if obj.tides is not None:
                # obj.logger.debug('Adding tides to hotstart_domain.')
                hotstart_domain.add_boundary_condition(obj.tides)
            if obj.nws2 is not None:
                # obj.logger.debug('Adding NWS2 object to hotstart_domain.')
                hotstart_domain.set_atmospheric_forcing(obj.nws2)
            if obj.hydrology is not None:
                for hydrology in obj.hydrology:
                    # obj.logger.debug(
                    #     'Adding hydrology object to hotstart_domain.')
                    hotstart_domain.add_hydrology(hydrology)
            obj.__dict__['hotstart_domain'] = hotstart_domain
        return hotstart_domain


class HotstartDriver:

    # reusing a descriptor from the schout module.
    surface_output_vars = SurfaceOutputVars()

    def __get__(self, obj, val):
        hotstart_driver = obj.__dict__.get('hotstart_driver')
        if hotstart_driver is None:
            # logger.debug('Creating hotstart driver.')
            hotstart_driver = ModelDriver(
                model_domain=obj.hotstart_domain,
                dt=obj.args.timestep,
                rnday=obj.args.forecast_days,
                # ihfskip=obj.args.ihfskip,
                start_date=obj.target_datetime,
                # ibc=obj.ibc,
                # stations=obj.stations,
                nspool=obj.nspool,
                combine_hotstart=obj.previous_run_directory / 'outputs',
                server_config=obj.server_config,
                **self.surface_outputs(obj)
                )
            obj.__dict__['hotstart_driver'] = hotstart_driver
        return hotstart_driver

    def surface_outputs(self, obj):
        surface_outputs = {}
        outvars = []
        for vardata in self.surface_output_vars.values():
            for varname, _ in vardata:
                outvars.append(varname)
        for key, val in obj.args.__dict__.items():
            if key in outvars and val is True:
                surface_outputs[key] = val
        return surface_outputs


class ForecastsDirectory:

    def __get__(self, obj, val):
        forecasts_directory = obj.__dict__.get('forecasts_directory')
        if forecasts_directory is None:
            forecasts_directory = obj.project_directory / 'forecasts'
            forecasts_directory.mkdir(exist_ok=True)
            obj.__dict__['forecasts_directory'] = forecasts_directory
        return forecasts_directory


class Forcings:

    def __get__(self, obj, val):
        forcings = obj.__dict__.get('forcings')
        if forcings is None:
            forcings = namedtuple(
                'forcings', ['tides', 'atmosphere', 'hydrology'])
            forcings = forcings(
                tides=obj.tides,
                atmosphere=obj.nws2,
                hydrology=obj.hydrology
                )
            obj.__dict__['forcings'] = forcings
        return forcings


class NWS2Descriptor:

    def __get__(self, obj, val):
        nws2 = obj.__dict__.get('nws2')
        if nws2 is None:
            sflux_1 = self.sflux_1(obj)
            if sflux_1 is not None:
                nws2 = NWS2(
                    sflux_1,
                    self.sflux_2(obj)
                    )
                obj.__dict__['nws2'] = nws2
        return nws2

    def sflux_1(self, obj):
        for arg, value in obj.args.__dict__.items():
            if 'gdas' in arg:
                if value is True:
                    raise NotImplementedError(
                        'GDAS product not implemented.')
            elif 'gfs' in arg:
                if value is True:
                    if arg == 'gfs':
                        return GFS()
                    else:
                        return GFS(product=arg)

    def sflux_2(self, obj):
        return


class Sflux2:

    def __get__(self, obj, val):
        sflux_1 = obj.__dict__.get('sflux_1')
        if sflux_1 is None:
            for arg, value in obj.args.__dict__.items():
                if 'gdas' in arg:
                    if value is True:
                        raise NotImplementedError(
                            'GDAS product not implemented.')
                elif 'gfs' in arg:
                    if value is True:
                        if arg == 'gfs':
                            sflux_1 = GFS()
                        else:
                            sflux_1 = GFS(product=arg)
            obj.__dict__['sflux_1'] = sflux_1
        return sflux_1


class HydrologyDescriptor:

    def __get__(self, obj, val):
        hydrology = obj.__dict__.get('hydrology')
        if hydrology is None:
            hydrology = []
            for hydro in obj.args.hydrology:
                if hydro == "NWM":
                    # obj.logger.debug('Append NWM object.')
                    hydrology.append(NWM())
            obj.__dict__['hydrology'] = hydrology
        return hydrology


class WindrotPath:
    def __get__(self, obj, val):
        windrot_path = obj.__dict__.get('windrot_path')
        if windrot_path is None:
            windrot_path = obj.static_files_directory / 'windrot_geo2proj.gr3'
            if obj.args.overwrite is True and windrot_path.exists():
                windrot_path.unlink()
            if not windrot_path.exists():
                obj.hotstart_domain.nws._windrot = obj.hotstart_domain.hgrid
                obj.hotstart_domain.nws._windrot.write(
                    windrot_path, overwrite=obj.args.overwrite)
            obj.__dict__['windrot_path'] = windrot_path
        return windrot_path


class ForecastUpdate(ForecastInit):

    forecasts_directory = ForecastsDirectory()
    hotstart_directory = HotstartDirectory()
    hotstart_domain = HotstartDomain()
    hotstart_driver = HotstartDriver()
    forcings = Forcings()
    nws2 = NWS2Descriptor()
    hydrology = HydrologyDescriptor()
    windrot_path = WindrotPath()

    def __init__(self, args: Namespace):
        self._args = args
        # logger.info("Starting forecast update sequence.")
        self._load_config()

        """
        TODO: Generate slurm-aware pre-processing of inputs.
        This means we need to create "from_file" constructors for each of the
        inputs.
        Essentially, if we are running this script on the "bare metal" all
        processes have to be run sequentially because we don't have enough
        CPU's. In contrast, with slurm and other process managers, we can
        launch multiple concurrent parallel pools.
        The current code base assumes bare metal processing.
        """
        self._symlink_files(
            self.hotstart_directory,
            self.hotstart_domain.ics,
            )
        if isinstance(self.forcings.atmosphere, NWS2):
            self._symlink_windrot(self.hotstart_directory)

        # logger.info("Calling hotstart write sequence.")
        self.hotstart_driver.write(
            self.hotstart_directory,
            hgrid=False,
            vgrid=False,
            fgrid=False,
            wind_rot=False,
            overwrite=self.args.overwrite
        )
        if self.args.skip_run is False:
            # self.logger.info("Executing SCHISM.")
            subprocess.check_call(
                ["make", "run"],
                cwd=self.hotstart_directory
            )

    def _load_config(self):
        config = self.project_directory / 'config.json'
        with open(config) as json_file:
            config = json.load(json_file)
        config.update(self.args.__dict__)
        self._args = Namespace(**config)

    def _symlink_windrot(self, target_directory):
        windrot_lnk = target_directory / 'windrot_geo2proj.gr3'
        if self.args.overwrite is True:
            if windrot_lnk.exists():
                windrot_lnk.unlink()
        os.symlink(
            os.path.relpath(
                self.windrot_path, target_directory), windrot_lnk)

    @property
    def nspool(self):
        if self.args.nspool is not None:
            if '.' in self.args.nspool:
                return timedelta(hours=float(self.args.nspool))
            else:
                return int(self.args.nspool)

    @property
    def previous_run_directory(self):
        if (self.coldstart_directory / 'outputs').exists():
            return self.coldstart_directory
        else:
            raise NotImplementedError(
                'Must find previous_run_directory from the hotstarts or '
                'regenerate init')



#! /usr/bin/env python
import unittest
from unittest.mock import patch
import sys
import tempfile
import warnings
import pathlib
from pyschism.cmd import plot_mesh


class PlotMeshCmdTestCase(unittest.TestCase):

    @patch('matplotlib.pyplot.show')
    def test_plot_mesh_command(self, mock):
        tmpdir = tempfile.TemporaryDirectory()
        outdir = pathlib.Path(tmpdir.name)
        nodes = {
            0: (0., 0., -1.),
            1: (1., 0., 0.),
            2: (1., 1., 0.),
            3: (0., 1., 0.),
            4: (0.5, 1.5, 0.),
        }
        elements = {
            0: [2, 4, 3],
            1: [0, 1, 2, 3],
        }
        with open(outdir / 'hgrid.gr3', 'w') as f:
            f.write('\n')
            f.write(f'{len(elements):d} ')
            f.write(f'{len(nodes):d}\n')
            for id, (x, y, z) in nodes.items():
                f.write(f"{id} ")
                f.write(f"{x} ")
                f.write(f"{y} ")
                f.write(f"{z}\n")
            for id, geom in elements.items():
                f.write(f"{id} ")
                f.write(f"{len(geom)} ")
                for idx in geom:
                    f.write(f"{idx:d} ")
                f.write(f"\n")
        cmd = ["plot_mesh"]
        cmd += [str(outdir / 'hgrid.gr3')]
        cmd += ["--plot-elements"]
        cmd += ["--plot-boundaries"]
        cmd += [f"--save-path={str(outdir / 'hgrid.png')}"]
        with patch.object(sys, 'argv', cmd):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self.assertEqual(plot_mesh.main(), 0)

    def test_init(self):
        with patch.object(plot_mesh, "main", return_value=0):
            with patch.object(plot_mesh, "__name__", "__main__"):
                with patch.object(plot_mesh.sys, 'exit') as mock_exit:
                    plot_mesh.init()
                    assert mock_exit.call_args[0][0] == 0


if __name__ == '__main__':
    unittest.main()



from functools import lru_cache
import os
from typing import Union, List

import numpy as np  # type: ignore[import]
from pyproj import CRS  # type: ignore[import]

from pyschism.forcing.tides.bctypes import BoundaryCondition
from pyschism.forcing.tides.tides import Tides
from pyschism.forcing.atmosphere.nws import NWS
from pyschism.forcing.hydrology.base import Hydrology
# from pyschism.forcing.atmosphere.nws.nws2 import NWS2
from pyschism.mesh import Hgrid, Vgrid, Fgrid
from pyschism.enums import Coriolis


# class HgridDescriptor:

#     def __set__(self, obj, hgrid: Hgrid):
#         if not isinstance(hgrid, Hgrid):
#             raise TypeError(f'Argument hgrid must be of type {Hgrid}, '
#                             f'not {type(hgrid)}')
#         obj.__dict__['hgrid'] = hgrid

#     def __get__(self, obj, val):
#         return obj.__dict__['hgrid']


# class VgridDescriptor:

#     def __set__(self, obj, vgrid: Union[Vgrid, None]):
#         if not isinstance(vgrid, (Vgrid, type(None))):
#             raise TypeError(f'Argument vgrid must be of type {Vgrid} or None, '
#                             f'not {type(vgrid)}')
#         if vgrid is None:
#             vgrid = Vgrid()
#         obj.__dict__['vgrid'] = vgrid

#     def __get__(self, obj, val):
#         return obj.__dict__['vgrid']


# class FgridDescriptor:

#     def __set__(self, obj, fgrid: Fgrid):
#         if not isinstance(fgrid, Fgrid):
#             raise TypeError(f'Argument fgrid must be of type {Fgrid} or None, '
#                             f'not {type(fgrid)}')
#         obj.__dict__['fgrid'] = fgrid

#     def __get__(self, obj, val):
#         return obj.__dict__['fgrid']


# class OpenBoundariesDescriptor:

#     def __get__(self, obj, val):
#         open_boudaries = obj.__dict__.get('open_boudaries')
#         if open_boudaries is None:
#             open_boudaries = {}
#             for bnd in obj.hgrid.boundaries.ocean.itertuples():
#                 open_boudaries[bnd.id] = {
#                     'indexes': bnd.indexes, 'forcing': None}
#             obj.__dict__['open_boudaries'] = open_boudaries
#         return open_boudaries


# class NwsDescriptor:

#     def __set__(self, obj, nws: NWS):
#         if not isinstance(nws, NWS):
#             raise TypeError(f"Argument nws must be of type {NWS}, not "
#                             f"type {type(nws)}.")
#         obj.__dict__['nws'] = nws

#     def __get__(self, obj, val):
#         return obj.__dict__.get('nws')




# class NcorDescriptor:

#     def __set__(self, obj, ncor: Coriolis):
#         if not isinstance(ncor, Coriolis):
#             raise TypeError(f"ncor must be of type {Coriolis}, not type "
#                             f"{type(ncor)}.")
#         if ncor == Coriolis.AUTO:
#             if obj.hgrid.ics == 1:
#                 obj.sfea0 = np.median(obj.hgrid.get_y("EPSG:4326"))
#             elif obj.hgrid.ics == 2:
#                 pass  # nothing to do for ics=2
#             else:
#                 raise ValueError(
#                     f'Unknown hgrid.ics parameter {obj.hgrid.ics}')

#         elif ncor == Coriolis.CORICOEFF:
#             obj.coricoef = 0.

#         elif ncor == Coriolis.RLATITUDE:
#             obj.rlatitude = 46.

#         else:
#             raise NotImplementedError(
#                 f"Unknown value for Coriolis enum type {ncor}.")
#         obj.__dict__['ncor'] = ncor

#     def __get__(self, obj, val):
#         return obj.__dict__.get('ncor', Coriolis.AUTO)


# class Sfea0Descriptor:

#     def __set__(self, obj, sfea0: float):
#         obj.__dict__['sfea0'] = sfea0

#     def __get__(self, obj, val):
#         return obj.__dict__.get('sfea0')


# class CoricoeffDescriptor:

#     def __set__(self, obj, coricoeff: float):
#         obj.__dict__['coricoeff'] = coricoeff

#     def __get__(self, obj, val):
#         return obj.__dict__.get('coricoeff', 0.)


# class RlatitudeDescriptor:

#     def __set__(self, obj, rlatitude: float):
#         obj.__dict__['rlatitude'] = rlatitude

#     def __get__(self, obj, val):
#         return obj.__dict__.get('rlatitude', 46.)


class OpenBoundaries:

    def __init__(self, hgrid: Hgrid):
        open_boundaries = {}
        for bnd in hgrid.boundaries.ocean().itertuples():
            open_boundaries[bnd.id] = {
                'indexes': bnd.indexes, 'forcing': None}
        self._hgrid = hgrid
        self._open_boundaries = open_boundaries

    def __call__(self):
        return self._open_boundaries

    def __len__(self):
        return len(self._hgrid.boundaries.ocean())

    def __getitem__(self, id):
        return self._open_boundaries[id]

    def __iter__(self):
        for id, data in self._open_boundaries.items():
            yield id, data


class ModelDomain:

    # _hgrid = HgridDescriptor()
    # _vgrid = VgridDescriptor()
    # _fgrid = FgridDescriptor()
    # _open_boundaries = OpenBoundariesDescriptor()
    # _nws = NwsDescriptor()
    # _ncor = NcorDescriptor()
    # _ics = IcsDescriptor()
    # sfea0 = Sfea0Descriptor()
    # coricoef = CoricoeffDescriptor()
    # rlatitude = RlatitudeDescriptor()

    def __init__(self, hgrid: Hgrid, vgrid: Vgrid, fgrid: Fgrid):
        """Class representing a SCHISM computational domain.

        This class combines the horizontal grid (hgrid), vertical grid (vgrid)
        and friction/drag grids (fgrid). Additionally, this class holds
        information about forcings.
        Args:
            hgrid: :class:`pyschism.mesh.Hgrid` instance.
            vgrid: :class:`pyschism.mesh.Vgrid` instance.
            fgrid: :class:`pyschism.mesh.Fgrid` derived instance.
        """
        self._hgrid = hgrid
        self._vgrid = vgrid
        self._fgrid = fgrid
        self._open_boundaries = OpenBoundaries(hgrid)
        self._ncor = Coriolis.AUTO
        self._nws: Union[NWS, None] = None
        self._hydrology: List[Hydrology] = []

    @staticmethod
    def open(hgrid: Union[str, os.PathLike], fgrid: Union[str, os.PathLike],
             vgrid: os.PathLike = None, hgrid_crs: Union[str, CRS] = None,
             fgrid_crs: Union[str, CRS] = None):
        """Open files from disk"""
        return ModelDomain(
            Hgrid.open(hgrid, hgrid_crs),
            Vgrid.open(vgrid) if vgrid is not None else Vgrid(),
            Fgrid.open(fgrid, fgrid_crs))

    def add_boundary_condition(self, forcing: BoundaryCondition, id=None):
        if id is None:
            for i in range(len(self.open_boundaries)):
                self.add_boundary_condition(forcing, i)
        else:
            if not isinstance(forcing, BoundaryCondition):
                raise TypeError("Argument must be of type "
                                f"{BoundaryCondition} but got type "
                                f"{type(forcing)}")
            self.open_boundaries[id]['forcing'] = forcing

    def set_atmospheric_forcing(self, atmospheric_forcing: NWS):
        self._nws = atmospheric_forcing

    def set_coriolis(self, ncor: Coriolis):
        self._ncor = ncor

    def add_hydrology(self, hydrology: Hydrology):
        assert isinstance(hydrology, Hydrology), \
            f"Argument hydrology must be of type {Hydrology}, " \
            f"not type {type(hydrology)}."
        self._hydrology.append(hydrology)

    @lru_cache(maxsize=1)
    def get_active_potential_constituents(self):
        # PySCHISM allows the user to input the tidal potentials individually
        # for each boundary, however, SCHISM supports only a global
        # specification. Here, we collect all the activated tidal potentials
        # on each boundary and activate them all globally
        const = dict()
        for id in self.open_boundaries():
            forcing = self.open_boundaries[id]['forcing']
            if isinstance(forcing, Tides):
                for active in forcing.get_active_potential_constituents():
                    const[active] = True
        return tuple(const.keys())

    @lru_cache(maxsize=1)
    def get_active_forcing_constituents(self):
        # PySCHISM allows the user to input the tidal forcings individually
        # for each boundary, however, SCHISM supports only a global
        # specification. Here, we collect all the activated tidal forcings
        # on each boundary and activate them all globally
        const = dict()
        for id in self.open_boundaries():
            forcing = self.open_boundaries[id]['forcing']
            if isinstance(forcing, Tides):
                for active in forcing.get_active_forcing_constituents():
                    const[active] = True
        return tuple(const.keys())

    def make_plot(self, **kwargs):
        if self.vgrid.is3D():
            raise NotImplementedError(
                "Plotting not yet supported for 3D meshes.")
        elif self.vgrid.is2D():
            self.hgrid.make_plot(**kwargs)

    @property
    def hgrid(self):
        return self._hgrid

    @property
    def vgrid(self):
        return self._vgrid

    @property
    def fgrid(self):
        return self._fgrid

    @property
    def nws(self):
        return self._nws

    @property
    def hydrology(self):
        return self._hydrology

    @property
    def ics(self):
        if self.hgrid.crs is None:
            return None
        elif self.hgrid.crs.is_geographic:
            return 2
        else:
            return 1

    @property
    def ncor(self):
        return self._ncor

    @property
    def open_boundaries(self):
        return self._open_boundaries

    @property
    def bctides(self):
        return self._bctides

    @property
    def _hgrid(self):
        return self.__hgrid

    @_hgrid.setter
    def _hgrid(self, hgrid: Hgrid):
        assert isinstance(hgrid, Hgrid), \
            f"Argument hgrid must be of type {Hgrid}, not type {type(hgrid)}."

        self.__hgrid = hgrid

    @property
    def _ncor(self):
        return self.__ncor

    @_ncor.setter
    def _ncor(self, ncor: Coriolis):
        if not isinstance(ncor, Coriolis):
            raise TypeError(f"ncor must be of type {Coriolis}, not type "
                            f"{type(ncor)}.")
        if ncor == Coriolis.AUTO:
            if self.ics == 1:
                self.sfea0 = np.median(self.hgrid.get_y("EPSG:4326"))
            elif self.ics == 2:
                pass  # nothing to do for ics=2
            else:
                raise ValueError(
                    f'Unknown hgrid.ics parameter {self.ics}')

        elif ncor == Coriolis.CORICOEFF:
            self.coricoef = 0.

        elif ncor == Coriolis.RLATITUDE:
            self.rlatitude = 46.

        else:
            raise NotImplementedError(
                f"Unknown value for Coriolis enum type {ncor}.")
        self.__ncor = ncor



from abc import ABC, abstractmethod


class TidalDataProvider(ABC):

    @abstractmethod
    def get_elevation(self, constituent, vertices):
        """Returns amplitude and phase of constituent at given vertices."""

    @abstractmethod
    def get_velocity(self, constituent, vertices):
        """Returns amplitude and phase of constituent at given vertices."""

    @property
    @abstractmethod
    def constituents(self):
        """Returns list of constituents available in database."""



#! /usr/bin/env python
import pathlib
from datetime import datetime, timedelta, timezone
import logging
from time import time

from pyschism.mesh import Hgrid, Vgrid, Fgrid
from pyschism import ModelDomain, ModelDriver, Stations
from pyschism.forcing import Tides
from pyschism.forcing.atmosphere.nws.nws2 import NWS2
from pyschism.forcing.atmosphere.gfs import GlobalForecastSystem as GFS



SIMPLE_SLURM_DRIVER = """#!/bin/bash --login
#SBATCH -D .
#SBATCH -J schism-test
#SBATCH -A nosofs
#SBATCH --mail-type=all
#SBATCH --mail-user=jaime.calzada@noaa.gov
#SBATCH --output=slurm.log
#SBATCH -n 500
#SBATCH --time=02:00:00
#SBATCH --partition=orion

set -e

module load intel/2020 impi/2020 netcdf/4.7.2-parallel

PATH=$HOME/SCHISM/schism/build/bin:$PATH

main() {
  mkdir -p outputs
  time srun pschism_TVD-VL
}

main
"""

# https://eev.ee/blog/2012/05/23/python-faq-descriptors/

PARENT = pathlib.Path(__file__).parent

logging.basicConfig(level=logging.INFO)

_logger = logging.getLogger(__name__)

if __name__ == '__main__':
    # open gr3 file
    _logger.info('Reading hgrid file...')
    _tic = time()
    hgrid = Hgrid.open(PARENT / 'hgrid.gr3',crs='EPSG:4326')
    _logger.info(f'Reading hgrind file took {time()-_tic}.')

    vgrid = Vgrid()
    fgrid = Fgrid.open(PARENT / 'drag.gr3',crs='EPSG:4326')

    # setup model domain
    domain = ModelDomain(hgrid, vgrid, fgrid)

    # set tidal boundary conditions
    elevbc = Tides()
    elevbc.use_all()  # activate all forcing constituents

    # connect the boundary condition to the domain
    domain.add_boundary_condition(elevbc)

    sflux_1 = GFS()
    # sflux_2 = HWRF()

    atmos = NWS2(
             sflux_1,
    #         sflux_2
         )

    domain.set_atmospheric_forcing(atmos)

    #  ------ Param options
    # dt and rnday are required arguments

    # Use int or float for seconds, or timedelta objects for pythonic
    # specifications
    dt = timedelta(seconds=150.)
    rnday = timedelta(days=4.)

    # Use an integer for number of steps or a timedelta to approximate
    # number of steps internally based on timestep
    nspool = timedelta(minutes=30.)

    # The dramp and start_date are optional parameters.
    # start_date is required when using forcings that
    # require some natural dates, (e.g. tides, winds, T&S ...)
    # If using a forcing that requires a start date is invoked, and no start
    # date is provided, the driver will throw an exception.
    # tzinfo is optional, UTC assumed if not provided
    dramp = 0.1 * rnday
    dramp = timedelta(0.0)

    start_date = datetime.utcnow() - dramp

    # Now we add station outputs
#    stations = Stations.from_file(PARENT / 'station.in', timedelta(minutes=6.),
#                                  elev=True, u=True, v=True)

    # init the model driver
    driver = ModelDriver(domain, dt, rnday, dramp=dramp, start_date=start_date,
                         #stations=stations, nspool=nspool, elev=True)
                         nspool=nspool, elev=True)

    # Output requests, as well as most other namelist variables, can be
    # modified post instantiation through their corresponding properties.
    # For example, the following line activates the depth averaged horizontal
    # velocity output request:
    driver.param.schout.dahv = True

    # write files to disk
    outdir = pathlib.Path('staging')
    driver.write(outdir, overwrite=True)




import calendar
from datetime import datetime, timedelta, timezone
import json
from typing import Union

import numpy as np
import requests


class CoopsDataCollector:

    url = "https://tidesandcurrents.noaa.gov/api/datagetter?"

    def fetch(self, station_id: str,  variable: str, start_date: datetime,
              rndays: Union[float, timedelta], datum: str = 'NAVD',
              units: str = 'metric'):

        # check start_date
        if not isinstance(start_date, datetime):
            raise TypeError(f'start_date must be a {datetime} instance.')

        # make date GMT
        if start_date.tzinfo is not None and \
                start_date.tzinfo.utcoffset(start_date) is not None:
            tzinfo = start_date.tzinfo
            start_date = start_date.astimezone(timezone(timedelta(0)))
        else:
            tzinfo = False

        # check rndays
        if not isinstance(rndays, timedelta):
            rndays = timedelta(days=rndays)

        end_date = start_date + rndays

        # set datum
        if datum not in ['MHHW', 'MHW', 'MTL', 'MSL', 'MLW', 'MLLW', 'NAVD88',
                         'STND', 'NAVD']:
            raise AttributeError(f'Invalid datum: {datum}.')
        if datum == 'NAVD88':
            datum = 'NAVD'

        responses = list()
        for start_date, end_date in self._iter_datetime_segments(
                start_date, end_date):
            params = self._get_params(station_id, start_date, end_date, datum,
                                      units)
            try:
                r = requests.get(self.url, params=params, timeout=10.)
                r.raise_for_status()
            except requests.exceptions.HTTPError:
                pass  # this means no data avilable
            except requests.exceptions.ConnectionError as errc:
                print("Error Connecting:", errc)
            except requests.exceptions.Timeout as errt:
                print("Timeout Error:", errt)
            except requests.exceptions.RequestException as err:
                print("Unknown error.", err)
            responses.append(r)
        data = dict()
        data['datetime'] = list()
        data['values'] = list()
        for i, response in enumerate(responses):
            json_data = json.loads(response.text)
            if 'error' in json_data.keys():
                _start_date, _end_date = list(
                    self._get_datetime_segments(start_date, end_date))[i]
                data['datetime'].append(_start_date)
                data['values'].append(np.nan)
                data['datetime'].append(_end_date)
                data['values'].append(np.nan)
                continue
            if 'x' not in data.keys():
                data['x'] = float(json_data['metadata']['lon'])
            if 'y' not in data.keys():
                data['y'] = float(json_data['metadata']['lat'])
            if 'name' not in data.keys():
                data['name'] = json_data['metadata']['name']
            for _data in json_data['data']:
                data['datetime'].append(
                    datetime.strptime(_data['t'], '%Y-%m-%d %H:%M'))
                try:
                    data['values'].append(float(_data['v']))
                except ValueError:
                    data['values'].append(np.nan)
        if 'name' not in data.keys():
            data['name'] = ''
        if np.all(data['values'] == np.nan):
            return None
        # change datetimes back to original timezone if tz-aware
        if tzinfo:
            data['datetime'] = [
                dt.replace(tzinfo=timezone.utc).astimezone(tzinfo)
                for dt in data['datetime']]
        return data

    def _get_params(self, station_id, start_date, end_date, datum, units):
        params = {}
        params['station'] = station_id
        params['begin_date'] = start_date.strftime('%Y%m%d %H:%M')
        params['end_date'] = end_date.strftime('%Y%m%d %H:%M')
        params['product'] = 'water_level'
        params['datum'] = datum
        params['units'] = units
        params['time_zone'] = 'gmt'
        params['format'] = 'json'
        params['application'] = 'PySCHISM'
        return params

    def _iter_datetime_segments(self, start_date, end_date):
        """
        https://www.ianwootten.co.uk/2014/07/01/splitting-a-date-range-in-python/
        """

        def get_datespan(interval):
            start_epoch = calendar.timegm(start_date.timetuple())
            end_epoch = calendar.timegm(end_date.timetuple())
            date_diff = end_epoch - start_epoch
            step = date_diff / interval
            delta = timedelta(seconds=step)
            current_date = start_date
            while current_date + delta <= end_date:
                to_date = (current_date + delta)
                yield current_date, to_date
                current_date += delta

        segments = [(start_date, end_date)]
        interval = 2
        while np.any([(_end_date - _start_date).total_seconds()
                      > timedelta(days=31).total_seconds()
                      for _start_date, _end_date in segments]):
            segments = [(from_datetime, to_datetime)
                        for from_datetime, to_datetime
                        in get_datespan(interval)]
            interval += 1
        for _start_date, _end_date in segments:
            yield _start_date, _end_date



from abc import ABC, abstractmethod
from collections import defaultdict
# from enum import Enum
from datetime import datetime, timedelta
import logging
import os
import pathlib
from typing import Union, List
import warnings

import cf  # type: ignore[import]
import cftime
from netCDF4 import Dataset  # type: ignore[import]
import numpy as np  # type: ignore[import]
import pytz


_logger = logging.getLogger(__name__)

class FieldsDescriptor:

    def __set__(self, obj, fields: cf.FieldList):
        if len(fields) == 0:
            warnings.warn(f"No data with name {obj.name} exists on the input "
                          'dataset.')
        ordered_indexes = np.argsort(
            [field.construct('time').reference_datetime for field
             in fields])
        fields = cf.FieldList([fields[idx] for idx in ordered_indexes])
        obj.__dict__[obj.type] = fields

    def __get__(self, obj, val):
        return obj.__dict__[obj.type]


class Nxgrids:

    def __set__(self, obj, fields: cf.FieldList):
        if len(fields) == 0:
            raise ValueError("No data with name 'lon' exists on the input "
                             'dataset.')
        ordered_indexes = np.argsort(
            [field.construct('time').reference_datetime for field
             in fields])
        fields = cf.FieldList([fields[idx] for idx in ordered_indexes])
        obj.__dict__['nx_grids'] = fields

    def __get__(self, obj, val):
        return obj.__dict__['nx_grids']


class Nygrids:

    def __set__(self, obj, fields: cf.FieldList):
        if len(fields) == 0:
            raise ValueError("No data with name 'lat' exists on the input "
                             'dataset.')
        ordered_indexes = np.argsort(
            [field.construct('time').reference_datetime for field
             in fields])
        fields = cf.FieldList([fields[idx] for idx in ordered_indexes])
        obj.__dict__['ny_grids'] = fields

    def __get__(self, obj, val):
        return obj.__dict__['ny_grids']


class DatetimeArrays:

    def __get__(self, obj, val):
        for field in obj.fields:
            dt_array = []
            for dt in field.construct('time').datetime_array:
                if not isinstance(dt, cftime.DatetimeGregorian):
                    continue
                dt_array.append(pytz.timezone('UTC').localize(
                    datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute)))
            yield dt_array


class DatetimeArray:

    def __get__(self, obj, val):
        dt_array = set()
        for dt_array_x in obj.datetime_arrays:
            for dt in dt_array_x:
                dt_array.add(dt)
        return [dt for dt in sorted(dt_array)]


class ReferenceDatetimes:

    def __get__(self, obj, val):
        for field in obj.fields:
            dt = field.construct('time').reference_datetime
            yield pytz.timezone('UTC').localize(
                datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute))


class Variable:

    fields = FieldsDescriptor()
    reference_datetimes = ReferenceDatetimes()
    datetime_arrays = DatetimeArrays()
    datetime_array = DatetimeArray()

    def __init__(self, fields, var_type, name, long_name,
                 standard_name, units):
        self.type = var_type
        self.name = name
        self.nx_grids = fields.select_by_ncvar('lon')
        self.ny_grids = fields.select_by_ncvar('lat')
        self.fields = fields.select_by_ncvar(self.name)
        self.long_name = long_name
        self.standard_name = standard_name
        self.units = units

    def get_fields(self, start_date: datetime = None,
                   rnday: Union[float, int, timedelta] = None,
                   ) -> cf.FieldList:

        if start_date is None:
            start_date = self.datetime_array[0]

        if start_date < np.min(self.datetime_array) or \
                start_date > np.max(self.datetime_array):
            raise ValueError(f'Requested start date {start_date} is out of '
                             f'range with {np.min(self.datetime_array)} and '
                             f'{np.max(self.datetime_array)}')
        if rnday is None:
            rnday = np.max(self.datetime_array[:-2]) - start_date

        elif isinstance(rnday, (int, float)):
            rnday = timedelta(days=rnday)

        end_date = start_date + rnday  # type: ignore[operator]

        if end_date <= np.min(self.datetime_array) or \
                end_date >= np.max(self.datetime_array):
            rnday = end_date - start_date
            raise ValueError(f'Requested rnday {rnday} has an end date of '
                             f'{end_date} which is out of range with '
                             f'start_date={np.min(self.datetime_array)} and '
                             f'end_date={np.max(self.datetime_array)}')
        fields = []
        for i, datetime_array in reversed(
                list(enumerate(self.datetime_arrays))):
            field_start_date = np.min(datetime_array)
            if field_start_date >= start_date:
                fields.append(self.fields[i])
            else:
                fields.append(self.fields[i])
                break
        return cf.FieldList(fields)


class BaseComponent(ABC):

    name: str = ''

    def write(self, outdir: Union[str, os.PathLike], level: int,
              overwrite: bool = False, start_date: datetime = None,
              rnday: Union[float, int, timedelta] = None):

        assert level in [1, 2]
        outdir = pathlib.Path(outdir)

        if start_date is None:
            for vartype in self.var_types:
                variable = getattr(self, vartype)
                if start_date is None:
                    start_date = np.min(variable.datetime_array)

        if start_date is not None:
            # naive condition
            if start_date.tzinfo is None \
                    or start_date.tzinfo.utcoffset(start_date) is None:
                start_date = pytz.timezone('UTC').localize(start_date)
            timezone = start_date.tzinfo

        stacks = []
        for i, field in enumerate(getattr(
                self, self.var_types[0]).get_fields(start_date, rnday)):
            stacks.append(f"sflux_{self.name}_{level}.{i+1:04d}.nc")
        for i, filename in enumerate(stacks):
            with Dataset(outdir / filename, 'w',
                         format='NETCDF3_CLASSIC') as dst:
                dst.setncatts({"Conventions": "CF-1.0"})
                # dimensions
                variable = getattr(self, self.var_types[0])
                dst.createDimension('nx_grid',
                                    variable.nx_grids[0].shape[1])
                dst.createDimension('ny_grid',
                                    variable.ny_grids[0].shape[0])
                dst.createDimension('time', None)
                # variables
                # lon
                dst.createVariable('lon', 'f4', ('ny_grid', 'nx_grid'))
                dst['lon'].long_name = "Longitude"
                dst['lon'].standard_name = "longitude"
                dst['lon'].units = "degrees_east"
                dst['lon'][:] = variable.nx_grids[0]
                # lat
                dst.createVariable('lat', 'f4', ('ny_grid', 'nx_grid'))
                dst['lat'].long_name = "Latitude"
                dst['lat'].standard_name = "latitude"
                dst['lat'].units = "degrees_north"
                dst['lat'][:] = variable.ny_grids[0]
                nc_start_date = list(variable.reference_datetimes)[0]
                resolution = (
                    list(variable.datetime_arrays)[0][0]
                    - nc_start_date).total_seconds() / (60*60*24)
                nc_start_date = nc_start_date.astimezone(timezone)
                dst.createVariable('time', 'f4', ('time',))
                dst['time'].long_name = 'Time'
                dst['time'].standard_name = 'time'
                dst['time'].units = f'days since {nc_start_date.year}-' \
                                    f'{nc_start_date.month}-'\
                                    f'{nc_start_date.day} ' \
                                    f'{nc_start_date.hour:02d}:'\
                                    f'{nc_start_date.minute:02d} ' \
                                    f'{nc_start_date.tzinfo}'
                dst['time'].base_date = (
                    nc_start_date.year,
                    nc_start_date.month,
                    nc_start_date.day,
                    nc_start_date.hour)
                dst['time'][:] = np.arange(
                    resolution,
                    resolution*(variable.fields[0].shape[0]+1),
                    step=resolution)
                for vartype in self.var_types:
                    variable = getattr(self, vartype)
                    dst.createVariable(
                        variable.name, 'f4', ('time', 'ny_grid', 'nx_grid'))
                    for field in variable.get_fields(start_date, rnday):
                        dst[variable.name][:] = field
                        setattr(
                            dst[variable.name], "long_name",
                            variable.long_name)
                        setattr(
                            dst[variable.name], "standard_name",
                            variable.standard_name)
                        setattr(
                            dst[variable.name], "units",
                            variable.units)

    @property
    @abstractmethod
    def var_types(self):
        """
        air: ['prmsl', 'spfh', 'stmp', 'uwind', 'vwind']
        prc: ['prate']
        rad: ['dlwrf', 'dswrf']
        """


class AirComponent(BaseComponent):

    name = 'air'

    def __init__(self, fields: cf.FieldList, prmsl_name='prmsl',
                 spfh_name='spfh', stmp_name='stmp', uwind_name='uwind',
                 vwind_name='vwind'):
        self.prmsl = Variable(fields, 'prmsl', prmsl_name,
                              "Pressure reduced to MSL",
                              "air_pressure_at_sea_level", "Pa")
        self.spfh = Variable(fields, 'spfh', spfh_name,
                             "Surface Specific Humidity (2m AGL)",
                             "specific_humidity", '1')
        self.stmp = Variable(fields, 'stmp', stmp_name,
                             "Surface Air Temperature (2m AGL)",
                             "air_temperature", "K")
        self.uwind = Variable(fields, 'uwind', uwind_name,
                              "Surface Eastward Air Velocity (10m AGL)",
                              "eastward_wind", "m/s")
        self.vwind = Variable(fields, 'vwind', vwind_name,
                              "Surface Northward Air Velocity (10m AGL)",
                              "northward_wind", "m/s")

    @property
    def var_types(self):
        return ['prmsl', 'spfh', 'stmp', 'uwind', 'vwind']


class PrcComponent(BaseComponent):

    name = 'prc'

    def __init__(self, fields: cf.FieldList, prate_name='prate'):
        self.prate = Variable(fields, 'prate', prate_name,
                              "Surface Precipitation Rate",
                              "air_pressure_at_sea_level", "kg/m^2/s")

    @property
    def var_types(self):
        return ['prate']


class RadComponent(BaseComponent):

    name = 'rad'

    def __init__(self, fields: cf.FieldList, dlwrf_name='dlwrf',
                 dswrf_name='dswrf'):
        self.dlwrf = Variable(fields, 'dlwrf', dlwrf_name,
                              "Downward Long Wave Radiation Flux",
                              "surface_downwelling_longwave_flux_in_air",
                              "W/m^2")
        self.dswrf = Variable(fields, 'dswrf', dswrf_name,
                              "Downward Short Wave Radiation Flux",
                              "surface_downwelling_shortwave_flux_in_air",
                              "W/m^2")

    @property
    def var_types(self):
        return ['dlwrf', 'dswrf']


class Resource:

    def __set__(self, obj, resource: Union[str, os.PathLike]):
        obj.__dict__['resource'] = resource

    def __get__(self, obj, val):
        return obj.__dict__['resource']


class Fields:

    def __get__(self, obj, val):
        fields = obj.__dict__.get('fields')
        if fields is None:
            fields = cf.read(obj.resource, ignore_read_error=True)
            # check lon
            try:
                lon = fields.select_by_ncvar('lon')[0]
            except IndexError:
                raise ValueError(f"Resource {obj.resource} does not contain a "
                                 "'lon' variable.")
            if len(lon.get_data_axes()) != 2:
                raise ValueError("'lon' variable must be a 2-dimensional "
                                 "array")
            fnames = lon.get_filenames()
            lons = fields.select_by_ncvar('lon')
            _logger.info(f'fields.select_by_var() returned {lons}')
            for i in range(len(lons) - 1):
                if not (lon.array == lons[i+1].array).all():
                    raise ValueError(
                        "Invalid sflux dataset. Found two different 'lon' "
                        f"fields on files {fnames} and "
                        f'{lons[i+1].get_filenames()}')
            # check lat
            try:
                lat = fields.select_by_ncvar('lat')[0]
            except IndexError:
                raise ValueError(f"Resource {obj.resource} does not contain a "
                                 "'lat' variable.")
            if len(lat.get_data_axes()) != 2:
                raise ValueError("'lat' variable must be a 2-dimensional "
                                 "array")
            fnames = lat.get_filenames()
            lats = fields.select_by_ncvar('lat')
            for i in range(len(lats) - 1):
                if not (lat.array == lats[i+1].array).all():
                    raise ValueError(
                        "Invalid sflux dataset. Found two different 'lat' "
                        f"fields on files {fnames} and "
                        f'{lats[i+1].get_filenames()}')
            obj.__dict__['fields'] = fields
        return fields


class SfluxDataset:

    resource = Resource()
    fields = Fields()

    def __init__(self, resource: Union[str, os.PathLike], prmsl_name='prmsl',
                 spfh_name='spfh', stmp_name='stmp', uwind_name='uwind',
                 vwind_name='vwind', prate_name='prate', dlwrf_name='dlwrf',
                 dswrf_name='dswrf'):
        self.resource = resource
        self.air = AirComponent(self.fields, prmsl_name=prmsl_name,
                                spfh_name=spfh_name, stmp_name=stmp_name,
                                uwind_name=uwind_name, vwind_name=vwind_name)
        self.prc = PrcComponent(self.fields, prate_name=prate_name)
        self.rad = RadComponent(self.fields, dlwrf_name=dlwrf_name,
                                dswrf_name=dswrf_name)

    def __call__(self, model_driver):
        pass

    def write(self, outdir: Union[str, os.PathLike], level: int,
              overwrite: bool = False, start_date: datetime = None,
              rnday: Union[float, int, timedelta] = None):
        outdir = pathlib.Path(outdir)
        if outdir.name != 'sflux':
            outdir /= 'sflux'
        outdir.mkdir(exist_ok=True)
        self.air.write(outdir, level, overwrite, start_date, rnday)
        self.prc.write(outdir, level, overwrite, start_date, rnday)
        self.rad.write(outdir, level, overwrite, start_date, rnday)










# class ComponentDescriptor:

#     # file_names = ComponentFileNames()

#     def __init__(self, var_name: str, default_vars: List[str]):
#         self.default_name = default_name
#         setattr(self, f'{default_name}_name', default_name)

#     def __get__(self, obj, val):
#         component = obj.__dict__.get(self.default_name)
#         if component is None:
#             varname = getattr(self, f"{self.default_name}_name")
#             component = obj.fields.select_by_ncvar(varname)
#             if len(component) == 0:
#                 warnings.warn(f"No data with name {varname} exists "
#                               'on the input dataset.')
#             obj.__dict__[f"{self.default_name}"] = component
#         return component






    # def __init__(self):
    #     start_date = set()
    #     for vartype in self.var_types:
    #         start_date.add(getattr(self, vartype).start_date)
    #     if len(start_date) > 1:
    #         raise ValueError("dimension mistmatch between variables of the "
    #                          "same type.")


            # datetime_array = variable
            # start_date.add(np.min(
            # setattr(f"start_date = np.min(variable.datetime_array)



    # def __init__

    # def __init__(self, fields: cf.FieldList, **kwargs):
    #     self.fields = fields
    #     self.nc_date_ranges = []

    #     for var_name, value in kwargs.items():
    #         var = var_name.split('_name')[0]
    #         setattr(self, f'{var}.name', value)
    #     date_ranges = defaultdict(list)
    #     for var in self.variables:  # type: ignore[attr-defined]
    #         for field in getattr(self, var):
    #             time = field.construct('time')
    #             date_ranges[f"{var}"].append(time.datetime_array)
    #     self.nc_date_ranges = date_ranges[f"{var}"]
    #     # self.nc_reference_datetime = time.reference_datetime
    #     self.nc_start_date = self.nc_date_ranges[0][0]
    #     self.nc_end_date = self.nc_date_ranges[-1][-2]
    #     self.nc_time_resolution = self.nc_start_date - time.reference_datetime
    #     # self.nc_rnday = self.nc_date_ranges[-1][-2] - self.nc_start_date
    #     self.nx_grids = self.fields.select_by_ncvar('lon')
    #     self.ny_grids = self.fields.select_by_ncvar('lat')


        # if start_date is None:
        #     start_date = self.nc_start_date

        # if not isinstance(rnday, timedelta):
        #     if rnday is None:
        #         rnday = self.nc_rnday
        #     else:
        #         rnday = timedelta(days=rnday)

        # cfdata = getattr(self, f"{self.variables[0]}")  # type: ignore[attr-defined]  # noqa: E501
        # ordered_indexes = np.argsort(
        #     [field.construct('time').reference_datetime for field
        #      in cfdata])
        # start_index = None
        # for i, current_idx in enumerate(ordered_indexes):
        #     next_idx = ordered_indexes[(i+1) % len(ordered_indexes)]
        #     current_start_date = self.nc_date_ranges[current_idx][0]
        #     next_start_date = self.nc_date_ranges[next_idx][0]
        #     if next_start_date < current_start_date:
        #         break
        #     if start_date >= current_start_date \
        #             and next_start_date > start_date:
        #         start_index = current_idx
        #         break
        # if start_index is None:
        #     if len(ordered_indexes) == 1:
        #         if start_date < self.nc_start_date \
        #                 or start_date > self.nc_end_date:
        #             raise ValueError("Date out of range")
        #         start_index = 0
        #     else:
        #         raise ValueError(
        #             f'No data found for the specified start_date = {start_date}. '
        #             f'Earliest start_date must be > {self.nc_start_date}.')

        # end_date = start_date + rnday  # type: ignore[operator]
        # end_index = None
        # for i, current_idx in enumerate(reversed(ordered_indexes)):
        #     next_idx = ordered_indexes[(i+1) % len(ordered_indexes)]
        #     current_end_date = self.nc_date_ranges[current_idx][-2]
        #     next_end_date = self.nc_date_ranges[next_idx][-2]
        #     if end_date <= current_end_date \
        #             and next_end_date > end_date:
        #         end_index = current_idx
        #         break
        #     if next_end_date < current_end_date:
        #         end_index = current_idx
        # if end_index is None:
        #     end_index = np.max(ordered_indexes)

        # start_index = np.min(ordered_indexes)
        # end_index = np.max(ordered_indexes)

        # tzinfo = pytz.timezone('utc') if not hasattr(start_date, 'tzinfo') \
        #     else start_date.tzinfo

        # for i in range(start_index, end_index+1):



from enum import Enum


class Stratification(Enum):
    BAROCLINIC = 0
    BAROTROPIC = 1


class Coriolis(Enum):
    AUTO = 1
    CORICOEFF = 0
    RLATITUDE = -1


class IofHydroVariables(Enum):
    elev = "elev"
    air_pressure = "air_pressure"
    air_temperature = "air_temperature"
    specific_humidity = "specific_humidity"
    solar_radiation = "solar_radiation"
    sensible_flux = "sensible_flux"
    latent_heat = "latent_heat"
    upward_longwave = "upward_longwave"
    downward_longwave = "downward_longwave"
    total_heat_flux = "total_heat_flux"
    evaporation = "evaporation"
    precipitation = "precipitation"
    bottom_stress = "bottom_stress"
    wind_speed = "wind_speed"
    wind_stress = "wind_stress"
    dahv = "dahv"
    vertical_velocity = "vertical_velocity"
    temp = "temp"
    salt = "salt"
    water_density = "water_density"
    diffusivity = "diffusivity"
    viscosity = "viscosity"
    TKE = "TKE"
    mixing_length = "mixing_length"
    hvel = "hvel"
    hvel_side = "hvel_side"
    wvel_elem = "wvel_elem"
    temp_elem = "temp_elem"
    salt_elem = "salt_elem"
    pressure_gradient = "pressure_gradient"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid HYDRO output variable.")


class IofDvdVariables(Enum):
    DVD_1 = "DVD_1"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid DVD output variable.")


class IofWwmVariables(Enum):
    WWM_1 = "WWM_1"
    WWM_2 = "WWM_2"
    WWM_3 = "WWM_3"
    WWM_4 = "WWM_4"
    WWM_5 = "WWM_5"
    WWM_6 = "WWM_6"
    WWM_9 = "WWM_9"
    WWM_10 = "WWM_10"
    WWM_11 = "WWM_11"
    WWM_12 = "WWM_12"
    WWM_13 = "WWM_13"
    WWM_14 = "WWM_14"
    WWM_15 = "WWM_15"
    WWM_16 = "WWM_16"
    WWM_17 = "WWM_17"
    WWM_18 = "WWM_18"
    WWM_19 = "WWM_19"
    WWM_20 = "WWM_20"
    WWM_21 = "WWM_21"
    WWM_22 = "WWM_22"
    WWM_23 = "WWM_23"
    WWM_24 = "WWM_24"
    WWM_25 = "WWM_25"
    WWM_26 = "WWM_26"
    WWM_27 = "WWM_27"
    WWM_28 = "WWM_28"
    WWM_energy_dir = "WWM_energy_dir"
    wave_force = "wave_force"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid WWM output variable.")


class IofGenVariables(Enum):
    GEN_1 = "GEN_1"
    GEN_2 = "GEN_2"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid GEN output variable.")


class IofAgeVariables(Enum):
    AGE_1 = "AGE_1"
    AGE_2 = "AGE_2"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid AGE output variable.")


class IofSedVariables(Enum):
    SED_depth_change = "SED_depth_change"
    SED_D50 = "SED_D50"
    SED_bed_stress = "SED_bed_stress"
    SED_bed_roughness = "SED_bed_roughness"
    SED_TSC = "SED_TSC"
    bed_thickness = "bed_thickness"
    bed_age = "bed_age"
    z0st = "z0st"
    z0cr = "z0cr"
    z0sw = "z0sw"
    z0wr = "z0wr"
    SED3D_1 = "SED3D_1"
    SED_bdld_1 = "SED_bdld_1"
    SED_bedfrac_1 = "SED_bedfrac_1"
    SED3D_2 = "SED3D_2"
    SED_bdld_2 = "SED_bdld_2"
    SED_bedfrac_3 = "SED_bedfrac_3"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid SED output variable.")


class IofEcoVariables(Enum):
    ECO_1 = "ECO_1"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid ECO output variable.")


class IofIcmVariables(Enum):
    ICM_Chl = "ICM_Chl"
    ICM_pH = "ICM_pH"
    ICM_PrmPrdt = "ICM_PrmPrdt"
    ICM_DIN = "ICM_DIN"
    ICM_PON = "ICM_PON"
    ICM_SED_BENDOC = "ICM_SED_BENDOC"
    ICM_SED_BENNH4 = "ICM_SED_BENNH4"
    ICM_SED_BENNO3 = "ICM_SED_BENNO3"
    ICM_SED_BENPO4 = "ICM_SED_BENPO4"
    ICM_SED_BENCOD = "ICM_SED_BENCOD"
    ICM_SED_BENDO = "ICM_SED_BENDO"
    ICM_SED_BENSA = "ICM_SED_BENSA"
    ICM_lfsav = "ICM_lfsav"
    ICM_stsav = "ICM_stsav"
    ICM_rtsav = "ICM_rtsav"
    ICM_tlfsav = "ICM_tlfsav"
    ICM_tstsav = "ICM_tstsav"
    ICM_trtsav = "ICM_trtsav"
    ICM_hcansav = "ICM_hcansav"
    ICM_CNH4 = "ICM_CNH4"
    ICM_CNH3 = "ICM_CNH3"
    ICM_CPIP = "ICM_CPIP"
    ICM_CPOS = "ICM_CPOS"
    ICM_CCH4 = "ICM_CCH4"
    ICM_CSO4 = "ICM_CSO4"
    ICM_CH2S = "ICM_CH2S"
    ICM_SEDPON1 = "ICM_SEDPON1"
    ICM_SEDPON2 = "ICM_SEDPON2"
    ICM_SEDPON3 = "ICM_SEDPON3"
    ICM_SEDPOP1 = "ICM_SEDPOP1"
    ICM_SEDPOP2 = "ICM_SEDPOP2"
    ICM_SEDPOP3 = "ICM_SEDPOP3"
    ICM_SEDPOC1 = "ICM_SEDPOC1"
    ICM_SEDPOC2 = "ICM_SEDPOC2"
    ICM_SEDPOC3 = "ICM_SEDPOC3"
    ICM_EROH2S = "ICM_EROH2S"
    ICM_EROLPOC = "ICM_EROLPOC"
    ICM_ERORPOC = "ICM_ERORPOC"
    ICM_DO_consumption = "ICM_DO_consumption"
    ICM_GP1 = "ICM_GP1"
    ICM_GP2 = "ICM_GP2"
    ICM_GP3 = "ICM_GP3"
    ICM_1 = "ICM_1"
    ICM_2 = "ICM_2"
    ICM_3 = "ICM_3"
    ICM_4 = "ICM_4"
    ICM_5 = "ICM_5"
    ICM_6 = "ICM_6"
    ICM_7 = "ICM_7"
    ICM_8 = "ICM_8"
    ICM_9 = "ICM_9"
    ICM_10 = "ICM_10"
    ICM_11 = "ICM_11"
    ICM_12 = "ICM_12"
    ICM_13 = "ICM_13"
    ICM_14 = "ICM_14"
    ICM_15 = "ICM_15"
    ICM_16 = "ICM_16"
    ICM_17 = "ICM_17"
    ICM_18 = "ICM_18"
    ICM_19 = "ICM_19"
    ICM_20 = "ICM_20"
    ICM_21 = "ICM_21"
    ICM_22 = "ICM_22"
    ICM_23 = "ICM_23"
    ICM_24 = "ICM_24"
    ICM_25 = "I6CM_25"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid ICM output variable.")


class IofCosVariables(Enum):
    COS_1 = "COS_1"
    COS_2 = "COS_2"
    COS_3 = "COS_3"
    COS_4 = "COS_4"
    COS_5 = "COS_5"
    COS_6 = "COS_6"
    COS_7 = "COS_7"
    COS_8 = "COS_8"
    COS_9 = "COS_9"
    COS_10 = "COS_10"
    COS_11 = "COS_11"
    COS_12 = "COS_12"
    COS_13 = "COS_13"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid COS output variable.")


class IofFibVariables(Enum):
    FIB_1 = "FIB_1"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid FIB output variable.")


class IofSed2dVariables(Enum):
    SED2D_depth_change = "SED2D_depth_change"
    SED2D_drag_coefficient = "SED2D_drag_coefficient"
    SED2D_cflsed = "SED2D_cflsed"
    SED2D_d50 = "SED2D_d50"
    SED2D_total_transport = "SED2D_total_transport"
    SED2D_susp_load = "SED2D_susp_load"
    SED2D_bed_load = "SED2D_bed_load"
    SED2D_average_transport = "SED2D_average_transport"
    SED2D_bottom_slope = "SED2D_bottom_slope"
    z0eq = "z0eq"
    z0cr2d = "z0cr2d"
    z0sw2d = "z0sw2d"
    z0wr2d = "z0wr2d"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid SED2D output variable.")


class IofMarshVariables(Enum):
    marsh_flag = "marsh_flag"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid MARSH output variable.")


class IofIceVariables(Enum):
    ICE_velocity = "ICE_velocity"
    ICE_strain_rate = "ICE_strain_rate"
    ICE_net_heat_flux = "ICE_net_heat_flux"
    ICE_fresh_water_flux = "ICE_fresh_water_flux"
    ICE_top_T = "ICE_top_T"
    ICE_tracer_1 = "ICE_tracer_1"
    ICE_tracer_2 = "ICE_tracer_2"
    ICE_tracer_3 = "ICE_tracer_3"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid ICE output variable.")


class IofAnaVariables(Enum):
    ANA_air_pres_grad_x = "ANA_air_pres_grad_x"
    ANA_air_pres_grad_y = "ANA_air_pres_grad_y"
    ANA_tide_pot_grad_x = "ANA_tide_pot_grad_x"
    ANA_tide_pot_grad_y = "ANA_tide_pot_grad_y"
    ANA_hor_viscosity_x = "ANA_hor_viscosity_x"
    ANA_hor_viscosity_y = "ANA_hor_viscosity_y"
    ANA_bclinic_force_x = "ANA_bclinic_force_x"
    ANA_bclinic_force_y = "ANA_bclinic_force_y"
    ANA_vert_viscosity_x = "ANA_vert_viscosity_x"
    ANA_vert_viscosity_y = "ANA_vert_viscosity_y"
    ANA_mom_advection_x = "ANA_mom_advection_x"
    ANA_mom_advection_y = "ANA_mom_advection_y"
    ANA_Richardson = "ANA_Richardson"
    ANA_transport_min_dt_elem = "ANA_transport_min_dt_elem"

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid ANA output variable.")


class StationOutputVariables(Enum):
    """Enumeration for stations output variables """
    ELEVATION = 'elev'
    AIR_PRESSURE = 'air_pressure'
    WINDX = 'windx'
    WINDY = 'windy'
    TEMPERATURE = 'T'
    SALINITY = 'S'
    U = 'u'
    V = 'v'
    W = 'w'

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid station output variable.")


class StationOutputIndex(Enum):
    """Indexing for stations output variables """
    ELEVATION = 0
    AIR_PRESSURE = 1
    WINDX = 2
    WINDY = 3
    TEMPERATURE = 4
    SALINITY = 5
    U = 6
    V = 7
    W = 8


class OutputVariableUnit(Enum):
    ELEVATION = 'meters'


class NWSType(Enum):
    """Atmospheric forcing type required by param.nml"""
    TIME_HISTORY = 1
    CLIMATE_AND_FORECAST = 2
    HEAT_CONSERVATION = 3
    UNSTRUCTURED = 4

    @classmethod
    def _missing_(self, name):
        raise ValueError(f"{name} is not a valid NWS type.")


class ForecastProduct(Enum):
    GDAS = 'gdas_0p25'
    GDAS_0P25 = 'gdas_0p25'
    GFS = 'gfs_0p25_1hr'
    GFS_0P25 = 'gfs_0p25'
    GFS_0P25_1HR = 'gfs_0p25_1hr'
    GFS_0P50 = 'gfs_0p50'
    GFS_1P00 = 'gfs_1p00'

    @classmethod
    def _missing_(self, name):
        ValueError(f'{name} is not a known atmospheric forecast product for '
                   'air.')


class GFSProduct(Enum):
    GFS_0P25 = 'gfs_0p25'
    GFS_0P25_1HR = 'gfs_0p25_1hr'
    GFS_0P50 = 'gfs_0p50'
    GFS_1P00 = 'gfs_1p00'


class ForecastCycle(Enum):
    z00 = '00z'
    z01 = '01z'
    z02 = '02z'
    z03 = '03z'
    z04 = '04z'
    z05 = '05z'
    z06 = '06z'
    z07 = '07z'
    z08 = '08z'
    z09 = '09z'
    z10 = '10z'
    z11 = '11z'
    z12 = '12z'
    z13 = '13z'
    z14 = '14z'
    z15 = '15z'
    z16 = '16z'
    z17 = '17z'
    z18 = '18z'
    z19 = '19z'
    z20 = '20z'
    z21 = '21z'
    z22 = '22z'
    z23 = '23z'

    @classmethod
    def _missing_(self, name):
        raise ValueError(f"{name} is not a valid cycle.")


class iof_hydro(Enum):
    elev = 0
    air_pressure = 1
    air_temperature = 2
    specific_humidity = 3
    solar_radiation = 4
    sensible_flux = 5
    latent_heat = 6
    upward_longwave = 7
    downward_longwave = 8
    total_heat_flux = 9
    evaporation = 10
    precipitation = 11
    bottom_stress = 12
    wind_speed = 13
    wind_stress = 14
    dahv = 15
    vertical_velocity = 16
    temp = 17
    salt = 18
    water_density = 19
    diffusivity = 20
    viscosity = 21
    TKE = 22
    mixing_length = 23
    hvel = 24
    hvel_side = 25
    wvel_elem = 26
    temp_elem = 27
    salt_elem = 28
    pressure_gradient = 29


class iof_dvd(Enum):
    DVD_1 = 0


class iof_wwm(Enum):
    WWM_1 = 0
    WWM_2 = 1
    WWM_3 = 2
    WWM_4 = 3
    WWM_5 = 4
    WWM_6 = 5
    WWM_9 = 6
    WWM_10 = 7
    WWM_11 = 8
    WWM_12 = 9
    WWM_13 = 10
    WWM_14 = 11
    WWM_15 = 12
    WWM_16 = 13
    WWM_17 = 14
    WWM_18 = 15
    WWM_19 = 16
    WWM_20 = 17
    WWM_21 = 18
    WWM_22 = 19
    WWM_23 = 20
    WWM_24 = 21
    WWM_25 = 22
    WWM_26 = 23
    WWM_27 = 24
    WWM_28 = 25
    WWM_energy_dir = 26
    wave_force = 27


class iof_gen(Enum):
    GEN_1 = 0
    GEN_2 = 1


class iof_age(Enum):
    AGE_1 = 0
    AGE_2 = 1


class iof_sed(Enum):
    SED_depth_change = 0
    SED_D50 = 1
    SED_bed_stress = 2
    SED_bed_roughness = 3
    SED_TSC = 4
    bed_thickness = 5
    bed_age = 6
    z0st = 7
    z0cr = 8
    z0sw = 9
    z0wr = 10
    SED3D_1 = 11
    SED_bdld_1 = 12
    SED_bedfrac_1 = 13
    SED3D_2 = 14
    SED_bdld_2 = 15
    SED_bedfrac_3 = 16


class iof_eco(Enum):
    ECO_1 = 0


class iof_icm(Enum):
    ICM_Chl = 0
    ICM_pH = 1
    ICM_PrmPrdt = 2
    ICM_DIN = 3
    ICM_PON = 4
    ICM_SED_BENDOC = 5
    ICM_SED_BENNH4 = 6
    ICM_SED_BENNO3 = 7
    ICM_SED_BENPO4 = 8
    ICM_SED_BENCOD = 9
    ICM_SED_BENDO = 10
    ICM_SED_BENSA = 11
    ICM_lfsav = 12
    ICM_stsav = 13
    ICM_rtsav = 14
    ICM_tlfsav = 15
    ICM_tstsav = 16
    ICM_trtsav = 17
    ICM_hcansav = 18
    ICM_CNH4 = 19
    ICM_CNH3 = 20
    ICM_CPIP = 21
    ICM_CPOS = 22
    ICM_CCH4 = 23
    ICM_CSO4 = 24
    ICM_CH2S = 25
    ICM_SEDPON1 = 26
    ICM_SEDPON2 = 27
    ICM_SEDPON3 = 28
    ICM_SEDPOP1 = 29
    ICM_SEDPOP2 = 30
    ICM_SEDPOP3 = 31
    ICM_SEDPOC1 = 32
    ICM_SEDPOC2 = 33
    ICM_SEDPOC3 = 34
    ICM_EROH2S = 35
    ICM_EROLPOC = 36
    ICM_ERORPOC = 37
    ICM_DO_consumption = 38
    ICM_GP1 = 39
    ICM_GP2 = 40
    ICM_GP3 = 41
    ICM_1 = 42
    ICM_2 = 43
    ICM_3 = 44
    ICM_4 = 45
    ICM_5 = 46
    ICM_6 = 47
    ICM_7 = 48
    ICM_8 = 49
    ICM_9 = 50
    ICM_10 = 51
    ICM_11 = 52
    ICM_12 = 53
    ICM_13 = 54
    ICM_14 = 55
    ICM_15 = 56
    ICM_16 = 57
    ICM_17 = 58
    ICM_18 = 59
    ICM_19 = 60
    ICM_20 = 61
    ICM_21 = 62
    ICM_22 = 63
    ICM_23 = 64
    ICM_24 = 65
    ICM_25 = 66


class iof_cos(Enum):
    COS_1 = 0
    COS_2 = 1
    COS_3 = 2
    COS_4 = 3
    COS_5 = 4
    COS_6 = 5
    COS_7 = 6
    COS_8 = 7
    COS_9 = 8
    COS_10 = 9
    COS_11 = 10
    COS_12 = 11
    COS_13 = 12


class iof_fib(Enum):
    FIB_1 = 0


class iof_sed2d(Enum):
    SED2D_depth_change = 0
    SED2D_drag_coefficient = 1
    SED2D_cflsed = 2
    SED2D_d50 = 3
    SED2D_total_transport = 4
    SED2D_susp_load = 5
    SED2D_bed_load = 6
    SED2D_average_transport = 7
    SED2D_bottom_slope = 8
    z0eq = 9
    z0cr2d = 10
    z0sw2d = 11
    z0wr2d = 12


class iof_marsh(Enum):
    marsh_flag = 0


class iof_ice(Enum):
    ICE_velocity = 0
    ICE_strain_rate = 1
    ICE_net_heat_flux = 2
    ICE_fresh_water_flux = 3
    ICE_top_T = 4
    ICE_tracer_1 = 5
    ICE_tracer_2 = 6
    ICE_tracer_3 = 7


class iof_ana(Enum):
    ANA_air_pres_grad_x = 0
    ANA_air_pres_grad_y = 1
    ANA_tide_pot_grad_x = 2
    ANA_tide_pot_grad_y = 3
    ANA_hor_viscosity_x = 4
    ANA_hor_viscosity_y = 5
    ANA_bclinic_force_x = 6
    ANA_bclinic_force_y = 7
    ANA_vert_viscosity_x = 8
    ANA_vert_viscosity_y = 9
    ANA_mom_advection_x = 10
    ANA_mom_advection_y = 11
    ANA_Richardson = 12
    ANA_transport_min_dt_elem = 13


class SchoutType(Enum):
    iof_hydro = iof_hydro
    iof_dvd = iof_dvd
    iof_wwm = iof_wwm
    iof_gen = iof_gen
    iof_age = iof_age
    iof_sed = iof_sed
    iof_eco = iof_eco
    iof_icm = iof_icm
    iof_cos = iof_cos
    iof_fib = iof_fib
    iof_sed2d = iof_sed2d
    iof_marsh = iof_marsh
    iof_ice = iof_ice
    iof_ana = iof_ana


class SchoutIofType(str, Enum):
    iof_hydro = 'iof_hydro'
    iof_dvd = 'iof_dvd'
    iof_wwm = 'iof_wwm'
    iof_gen = 'iof_gen'
    iof_age = 'iof_age'
    iof_sed = 'iof_sed'
    iof_eco = 'iof_eco'
    iof_icm = 'iof_icm'
    iof_cos = 'iof_cos'
    iof_fib = 'iof_fib'
    iof_sed2d = 'iof_sed2d'
    iof_marsh = 'iof_marsh'
    iof_ice = 'iof_ice'
    iof_ana = 'iof_ana'

    @classmethod
    def _missing_(self, name):
        raise ValueError(f"{name} is not a valid SCHISM output type.")


class OutputVariableShortName(Enum):

    ELEVATION = 'elev'

    @classmethod
    def _missing_(self, name):
        raise ValueError(
            f"{name} is not a valid SCHISM output variable short name")


class NationalWaterModelDataSource(Enum):
    AWS = 'AWS'
    FTP = 'FTP'
    NOMADS = 'NOMADS'

    @classmethod
    def _missing_(self, name):
        raise ValueError(
            f"{name} is not a valid National Water Model data source.")



from importlib import util
# import logging

from pyschism.param import Param
from pyschism.stations import Stations
from pyschism.domain import ModelDomain
from pyschism.driver import ModelDriver

__all__ = ["Param", "ModelDomain", 'ModelDriver', 'Stations']


if util.find_spec("colored_traceback") is not None:
    import colored_traceback  # type: ignore[import]
    colored_traceback.add_hook(always=True)

# logging.basicConfig(
#     level=logging.NOTSET,
#     format='[%(asctime)s] %(name)-13s %(levelname)-8s: %(message)s',
#     )
# logging.getLogger('matplotlib').setLevel(logging.WARNING)
# logging.getLogger('fiona').setLevel(logging.CRITICAL)
# logging.getLogger('rasterio').setLevel(logging.WARNING)



from enum import Enum
import os
import pathlib
from typing import Union

from pyproj import CRS  # type: ignore[import]

from pyschism.mesh.base import Gr3
from pyschism.mesh.parsers import grd


class FrictionFilename(Enum):
    MANNINGS_N = 'manning.gr3'
    DRAG_COEFFICIENT = 'drag.gr3'
    ROUGHNESS_LENGTH = 'rough.gr3'

    @classmethod
    def _missing_(self, name):
        raise ValueError(f'{name} is not a valid filename for a friction '
                         'file.')


class NchiType(Enum):
    MANNINGS_N = -1
    ROUGHNESS_LENGTH = 1
    DRAG_COEFFICIENT = 0


class Fgrid(Gr3):
    """
    Base class for all friction types (e.g. manning.grd, drag.grd, etc...)
    """

    def __init__(self, nchi: NchiType, *argv, **kwargs):
        self._nchi = nchi
        self._fname = FrictionFilename[NchiType(nchi).name]
        super().__init__(*argv, **kwargs)

    @property
    def nchi(self):
        return self._nchi.value

    @property
    def fname(self):
        return self._fname.value

    @staticmethod
    def open(file: Union[str, os.PathLike],
             crs: Union[str, CRS] = None):
        filename = pathlib.Path(file).name
        return FrictionDispatch[
            FrictionFilename(filename).name].value(
                **grd.read(pathlib.Path(file), boundaries=False, crs=crs))


class ManningsN(Fgrid):
    """  Class for representing Manning's n values.  """

    def __init__(self, *argv, **kwargs):
        self.hmin_man = 1.
        super().__init__(NchiType.MANNINGS_N, *argv, **kwargs)

    @classmethod
    def open(cls, file: Union[str, os.PathLike],
             crs: Union[str, CRS] = None):
        return super(Fgrid, cls).open(file, crs)


class RoughnessLength(Fgrid):

    def __init__(self, *argv, **kwargs):
        super().__init__(NchiType.ROUGHNESS_LENGTH, *argv, **kwargs)

    @classmethod
    def open(cls, file: Union[str, os.PathLike],
             crs: Union[str, CRS] = None):
        return super(Fgrid, cls).open(file, crs)


class DragCoefficient(Fgrid):

    def __init__(self, *argv, **kwargs):
        self.dzb_min = 0.5
        self.dzb_decay = 0.
        super().__init__(NchiType.DRAG_COEFFICIENT, *argv, **kwargs)

    @classmethod
    def open(cls, file: Union[str, os.PathLike],
             crs: Union[str, CRS] = None):
        return super(Fgrid, cls).open(file, crs)


class FrictionDispatch(Enum):
    MANNINGS_N = ManningsN
    DRAG_COEFFICIENT = DragCoefficient
    ROUGHNESS_LENGTH = RoughnessLength



from datetime import timedelta
import logging
from typing import Union

from pyschism.enums import (
    IofHydroVariables,
    IofDvdVariables,
    IofWwmVariables,
    IofGenVariables,
    IofAgeVariables,
    IofSedVariables,
    IofEcoVariables,
    IofIcmVariables,
    IofCosVariables,
    IofFibVariables,
    IofSed2dVariables,
    IofMarshVariables,
    IofIceVariables,
    IofAnaVariables,
    SchoutType
)


_logger = logging.getLogger(__name__)


class OutputVariableDescriptor:

    def __init__(self, iof_type, name, index):
        self._iof_type = iof_type
        self._name = name
        self._index = index

    def __get__(self, obj, val):
        return bool(getattr(obj, f'_{self._iof_type}')[self._index])

    def __set__(self, obj, val: bool):
        if not isinstance(val, bool):
            raise TypeError(f'Argument to {self._name} must be boolean, not '
                            f'type {type(val)}.')
        iof = getattr(obj, f'_{self._iof_type}')
        iof[self._index] = int(val)


class SurfaceOutputVars:

    def __init__(self):
        self._surface_output_vars = {
            'iof_hydro': [(var.value, i) for i, var
                          in enumerate(IofHydroVariables)],
            'iof_dvd': [(var.value, i) for i, var
                        in enumerate(IofDvdVariables)],
            'iof_wwm': [(var.value, i) for i, var
                        in enumerate(IofWwmVariables)],
            'iof_gen': [(var.value, i) for i, var
                        in enumerate(IofGenVariables)],
            'iof_age': [(var.value, i) for i, var
                        in enumerate(IofAgeVariables)],
            'iof_sed': [(var.value, i) for i, var
                        in enumerate(IofSedVariables)],
            'iof_eco': [(var.value, i) for i, var
                        in enumerate(IofEcoVariables)],
            'iof_icm': [(var.value, i) for i, var
                        in enumerate(IofIcmVariables)],
            'iof_cos': [(var.value, i) for i, var
                        in enumerate(IofCosVariables)],
            'iof_fib': [(var.value, i) for i, var
                        in enumerate(IofFibVariables)],
            'iof_sed2d': [(var.value, i) for i, var
                          in enumerate(IofSed2dVariables)],
            'iof_marsh': [(var.value, i) for i, var
                          in enumerate(IofMarshVariables)],
            'iof_ice': [(var.value, i) for i, var
                        in enumerate(IofIceVariables)],
            'iof_ana': [(var.value, i) for i, var
                        in enumerate(IofAnaVariables)],
        }

    def __get__(self, obj, val):
        return self._surface_output_vars


class SchoutMeta(type):

    surface_output_vars = SurfaceOutputVars()

    def __new__(meta, name, bases, attrs):
        for iof_type in meta.surface_output_vars.keys():
            attrs[f'_{iof_type}'] = len(SchoutType[iof_type].value)*[0]

        for iof_type, vardata in meta.surface_output_vars.items():
            for name, index in vardata:
                attrs[name] = OutputVariableDescriptor(iof_type, name, index)
        attrs['surface_output_vars'] = meta.surface_output_vars
        return type(name, bases, attrs)


class Nhot:

    def __set__(self, obj, nhot: int):
        if nhot not in [0, 1]:
            raise ValueError(f"nhot must be 0 or 1, not {nhot}")
        obj.__dict__['nhot'] = nhot

    def __get__(self, obj, val):
        return obj.__dict__.get('nhot')


class NhotWrite:

    def __set__(self, obj, nhot_write: int):
        obj.__dict__['nhot_write'] = nhot_write
        obj.__dict__['nhot'] = 1

    def __get__(self, obj, val):
        return obj.__dict__.get('nhot_write')


class IoutSta:

    def __set__(self, obj, iout_sta: int):
        if iout_sta not in [0, 1]:
            raise ValueError(f"iout_sta must be 0 or 1, not {iout_sta}")
        obj.__dict__['iout_sta'] = iout_sta

    def __get__(self, obj, val):
        return obj.__dict__.get('iout_sta')


class NspoolSta:

    def __set__(self, obj, nspool_sta: Union[int, timedelta]):
        obj.__dict__['nspool_sta'] = nspool_sta
        obj.__dict__['iout_sta'] = 1

    def __get__(self, obj, val):
        return obj.__dict__.get('nspool_sta')


class SCHOUT(metaclass=SchoutMeta):
    """ Provides error checking implementation for SCHOUT group """
    _iout_sta = IoutSta()
    _nhot = Nhot()
    nhot_write = NhotWrite()
    nspool_sta = NspoolSta()

    def __init__(self, **outputs):
        _logger.info('Initializing SCHOUT.')
        for key, val in outputs.items():
            setattr(self, key, val)

    def __iter__(self):
        for outvar in self._surface_output_vars:
            yield outvar, getattr(self, outvar)

    def __str__(self):
        schout = ["&SCHOUT"]
        if self.nhot_write is not None:
            schout.append(f"  nhot={self._nhot}")
            schout.append(f"  nhot_write={self.nhot_write}")
        if self.nspool_sta is not None:
            schout.append(f"  iout_sta={self._iout_sta}")
            schout.append(f"  nspool_sta={self.nspool_sta}")
        for var in dir(self):
            if var.startswith('_iof'):
                for i, state in enumerate(getattr(self, var)):
                    if state == 1:
                        schout.append(f'  {var[1:]}({i+1})={state}')
        schout.append('/')
        return '\n'.join(schout)



from abc import ABC, abstractmethod
from datetime import timedelta, timezone
from enum import Enum
from functools import lru_cache
import logging

from pyschism.domain import ModelDomain
from pyschism.forcing.tides.tides import Tides
from pyschism.param.param import Param

_logger = logging.getLogger(__name__)

class NullWritter:

    def __init__(self, *argv):
        pass

    def __str__(self):
        return ""


class NotImplementedWritter:

    def __init__(self, *argv):
        raise NotImplementedError('Writter for variable is not implemented.')


class TidalVariableWritter(ABC):

    def __init__(self, boundary, bctides):
        self.forcing = boundary['forcing']
        self.indexes = boundary['indexes']
        self.active_constituents = bctides.get_active_forcing_constituents()
        self._model_domain = bctides._model_domain

    @abstractmethod
    def __str__(self):
        raise NotImplementedError(f'str({self.__name__})')


class TidalElevationWritter(TidalVariableWritter):

    def __str__(self):
        f = ""
        for constituent in self.active_constituents:
            f += f'{constituent}\n'
            vertices = self._model_domain.hgrid.get_xy(
                crs='EPSG:4326')[self.indexes, :]
            amp, phase = self.forcing.get_elevation(constituent, vertices)
            for i in range(len(vertices)):
                f += f'{amp[i]:.8e} {phase[i]:.8e}\n'
        return f


class TidalVelocityWritter(TidalVariableWritter):

    def __str__(self):
        f = ''
        for constituent in self.active_constituents:
            f += f'{constituent}\n'
            vertices = self._model_domain.hgrid.get_xy(
                crs='EPSG:4326')[self.indexes, :]
            uamp, uphase, vamp, vphase = self.forcing.get_velocity(
                constituent, vertices)
            for i in range(len(vertices)):
                f += f'{uamp[i]:.8e} {uphase[i]:.8e} ' \
                     f'{vamp[i]:.8e} {vphase[i]:.8e}\n'
        return f


class iettypeWritter(Enum):
    NONE = NullWritter
    TIME_VARYING = NotImplementedWritter
    CONSTANT = NotImplementedWritter
    TIDAL = TidalElevationWritter
    SPACE_TIME_VARYING = NotImplementedWritter
    TIDAL_AND_SPACE_TIME_VARYING = NotImplementedWritter


class ifltypeWritter(Enum):
    NONE = NullWritter
    TIME_VARYING = NotImplementedWritter
    CONSTANT = NotImplementedWritter
    TIDAL = TidalVelocityWritter
    SPACE_TIME_VARYING = NotImplementedWritter
    TIDAL_AND_SPACE_TIME_VARYING = NotImplementedWritter


class itetypeWritter(Enum):
    NONE = NullWritter
    TIME_VARYING = NotImplementedWritter
    CONSTANT = NotImplementedWritter
    INITIAL_PROFILE_FOR_INFLOW = NotImplementedWritter
    INPUT_3D = NotImplementedWritter


class isatypeWritter(Enum):
    NONE = NullWritter
    TIME_VARYING = NotImplementedWritter
    CONSTANT = NotImplementedWritter
    INITIAL_PROFILE_FOR_INFLOW = NotImplementedWritter
    INPUT_3D = NotImplementedWritter


class itrtypeWritter(Enum):
    NONE = NullWritter
    TIME_VARYING = NotImplementedWritter
    CONSTANT = NotImplementedWritter
    INITIAL_PROFILE_FOR_INFLOW = NotImplementedWritter
    INPUT_3D = NotImplementedWritter


class Bctides:

    def __init__(self, model_domain: ModelDomain, param: Param,
                 cutoff_depth: float = 50.):
        """Provides an interface to write bctides.in to file. """
        _logger.info('Initializing Bctides.')
        # check if start_date was given in case tidal forcings are requested.
        # Note: This is done twice so that this class can be used independently
        # from Param to just write bctides files
        afc = model_domain.get_active_forcing_constituents()
        if len(afc) > 0 and param.opt.start_date is None:
            raise Exception('start_date argument is required for simulating '
                            'tidal forcing.')

        self._model_domain = model_domain
        self._param = param
        self._cutoff_depth = cutoff_depth

        # init the main tidal forcing object
        tides = Tides()
        for const in tides.all_constituents:
            tides.use_constituent(
                const,
                potential=True if const in
                self.get_active_potential_constituents() else False,
                forcing=True if const in
                self.get_active_forcing_constituents() else False
            )
        self.__tidal_forcing = tides

    def __str__(self):
        f = f"{self.start_date}\n" \
            f"{self.ntip} {self._cutoff_depth}\n"
        if self.ntip > 0:
            for constituent in self.get_active_potential_constituents():
                forcing = self.tidal_forcing(
                    self.start_date, self.rnday, constituent)
                f += f'{constituent} \n' \
                     f'{forcing[0]:G} ' \
                     f"{forcing[1]:G} " \
                     f'{forcing[2]:G} ' \
                     f'{forcing[3]:G} ' \
                     f'{forcing[4]:G}\n'
        f += f'{self.tidal_forcing.nbfr:d}\n'
        for constituent in self.get_active_forcing_constituents():
            forcing = self.tidal_forcing(
                self.start_date, self.rnday, constituent)
            f += f'{constituent} \n' \
                 f"{forcing[2]:G} " \
                 f'{forcing[3]:G} ' \
                 f'{forcing[4]:G}\n'
        f += f"{len(self._model_domain.open_boundaries)}\n"  # nope
        for id, data in self._model_domain.open_boundaries:
            f += f"{len(data['indexes'])} " \
                 f'{str(data["forcing"])}\n' \
                 f'{iettypeWritter[data["forcing"].iettype.name].value(data, self)}' \
                 f'{ifltypeWritter[data["forcing"].ifltype.name].value(data, self)}' \
                 f'{itetypeWritter[data["forcing"].itetype.name].value(data, self)}' \
                 f'{isatypeWritter[data["forcing"].isatype.name].value(data, self)}' \
                 f'{itrtypeWritter[data["forcing"].itrtype.name].value(data, self)}'
        return f

    @lru_cache(maxsize=1)
    def get_active_potential_constituents(self):
        # PySCHISM allows the user to input the tidal potentials and forcings
        # individually at each boundary, however, SCHISM supports only a global
        # specification. Here, we collect all the activated tidal potentials
        # on each boundary and activate them all globally
        # set active tidal potential constituents
        const = dict()
        for id, data in self._model_domain.open_boundaries:
            forcing = data['forcing']
            if isinstance(forcing, Tides):
                for active in forcing.get_active_potential_constituents():
                    const[active] = True
        return tuple(const.keys())

    @lru_cache(maxsize=1)
    def get_active_forcing_constituents(self):
        # set active tidal forcing constituents
        const = dict()
        for id, data in self._model_domain.open_boundaries:
            forcing = data['forcing']
            if isinstance(forcing, Tides):
                for active in forcing.get_active_forcing_constituents():
                    const[active] = True
        return tuple(const.keys())

    def write(self, path, overwrite=False):
        with open(path, 'w') as f:
            f.write(str(self))

    @property
    def start_date(self):
        return self._param.opt.start_date

    @property
    def rnday(self):
        return self._param.core.rnday

    @property
    def ntip(self):
        return len(self.get_active_potential_constituents())

    @property
    def tidal_forcing(self):
        return self.__tidal_forcing

    @property
    def start_date_utc(self):
        if self.start_date.tzinfo is not None and \
                self.start_date.tzinfo.utcoffset(self.start_date) is not None:
            return self.__start_date.astimezone(timezone(timedelta(0)))
        else:
            return self.__start_date
